package io.gitlab.jfronny.muscript.ast.dynamic;

import io.gitlab.jfronny.muscript.ast.DynamicExpr;
import io.gitlab.jfronny.muscript.annotations.CanThrow;
import io.gitlab.jfronny.muscript.annotations.UncheckedDynamic;
import io.gitlab.jfronny.muscript.data.Scope;
import io.gitlab.jfronny.muscript.data.dynamic.*;

import java.util.ArrayList;
import java.util.List;

@CanThrow
@UncheckedDynamic
public class Call extends DynamicExpr {
    private final DynamicExpr left;
    private final List<DynamicExpr> args;

    public Call(int chStart, int chEnd, DynamicExpr left, List<DynamicExpr> args) {
        super(chStart, chEnd);
        this.left = left;
        this.args = args;
    }

    @Override
    public Dynamic<?> get(Scope dataRoot) {
        try {
            return left.get(dataRoot).asCallable().getValue().apply(DFinal.of(args.stream().map(e -> e.get(dataRoot)).toArray(Dynamic[]::new)));
        } catch (DynamicTypeConversionException e) {
            throw e.locational(chStart, chEnd);
        }
    }

    @Override
    public DynamicExpr optimize() {
        DynamicExpr left = this.left.optimize();
        List<DynamicExpr> args = new ArrayList<>();
        for (DynamicExpr arg : this.args) args.add(arg.optimize());
        return new Call(chStart, chEnd, left, args);
    }

    @Override
    public boolean equals(Object obj) {
        return obj instanceof Call call && left.equals(call.left) && args.equals(call.args);
    }
}
