package io.gitlab.jfronny.muscript.ast.dynamic;

import io.gitlab.jfronny.muscript.ast.DynamicExpr;
import io.gitlab.jfronny.muscript.ast.Expr;
import io.gitlab.jfronny.muscript.data.Scope;
import io.gitlab.jfronny.muscript.data.dynamic.*;
import io.gitlab.jfronny.muscript.error.LocationalException;

import java.util.List;

public class Closure extends DynamicExpr {
    private final List<String> boundArgs;
    private final List<Expr<?>> steps;
    private final DynamicExpr fin;

    public Closure(int chStart, int chEnd, List<String> boundArgs, List<Expr<?>> expressions) {
        this(chStart, chEnd, boundArgs, expressions.subList(0, expressions.size() - 1), expressions.get(expressions.size() - 1).asDynamicExpr());
    }

    private Closure(int chStart, int chEnd, List<String> boundArgs, List<Expr<?>> steps, DynamicExpr fin) {
        super(chStart, chEnd);
        this.boundArgs = List.copyOf(boundArgs);
        this.steps = List.copyOf(steps);
        this.fin = fin;
    }

    @Override
    public DCallable get(Scope dataRoot) {
        return DFinal.of(args -> {
            if (args.size() != boundArgs.size()) throw new LocationalException(chStart, chEnd, "Invoked with unexpected number of parameters");
            Scope fork = dataRoot.fork();
            for (int i = 0; i < boundArgs.size(); i++) fork.set(boundArgs.get(i), args.get(i));
            for (Expr<?> step : steps) {
                step.get(fork);
            }
            return fin.get(fork);
        });
    }

    @Override
    public DynamicExpr optimize() {
        // Eliminating side effect free steps might be possible, but would be too much work imo
        return new Closure(chStart, chEnd, boundArgs, steps.stream().<Expr<?>>map(Expr::optimize).toList(), fin.optimize());
    }

    @Override
    public boolean equals(Object obj) {
        return obj instanceof Closure closure
                && boundArgs.equals(closure.boundArgs)
                && steps.equals(closure.steps)
                && fin.equals(closure.fin);
    }
}
